/*
  This file is under Public Domain.
*/
#ifdef __MINGW32__
#  define _WIN32_IE 0x0500
#endif

#include <windows.h>
#include <commctrl.h>
#include <stdlib.h>
#include <string.h>
#include <tchar.h>
#include <wchar.h>
#define KMELEON_PLUGIN_EXPORTS
#include <kmeleon_plugin.h>
#include <KMeleonConst.h>

#include <nsNetCID.h>
#include <nsStringAPI.h>
#include <nsEmbedString.h>
#include <nsAutoLock.h>
#include <nsIWebBrowser.h>
#include <nsIWebBrowserChrome.h>
#include <nsIWebNavigation.h>
#include <nsIDNSService.h>
#include <nsIDNSListener.h>
#include <nsIDNSRecord.h>
#include <nsIURI.h>
#include <nsICancelable.h>
#include <nsIEmbeddingSiteWindow.h>

#include "miniCOMPtr.h"
#include "ministring.h"

#ifndef WM_CLOSETAB
#define WM_CLOSETAB WM_APP + 150
#define WM_CREATETAB WM_APP + 151
#define WM_SWITCHTAB WM_APP + 152
#endif

#define PLUGIN_NAME "IPBar plugin"
#define TOOLBAND_NAME _T("IPBar")
#define BARPROPNAME TEXT("IPBarWindowPtr")

long DoMessage(const char *, const char *, const char *, long, long);
void Load();
void Create(HWND);
void Close(HWND);
void DoRebar(HWND);
int DoAccel(char *);

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

void DoGetList(char **);

void PostUpdateMessage();
BOOL DoUpdate(HWND);
HWND get_ipbar_hwnd(HWND);

int message_box(HWND, LPCSTR, LPCSTR, UINT);

WNDPROC g_KMeleonWndProc;
UINT g_id;

kmeleonPlugin kPlugin = {
	KMEL_PLUGIN_VER,
	PLUGIN_NAME,
	DoMessage
};

long DoMessage(const char *to, const char *from, const char *subject,
               long data1, long data2)
{
	if (to[0] == '*' || stricmp(to, kPlugin.dllname) == 0) {
		if (stricmp(subject, "Load") == 0) {
			Load();
		}
		else if (stricmp(subject, "Create") == 0) {
			Create((HWND) data1);
		}
		else if (stricmp(subject, "Close") == 0) {
			Close((HWND) data1);
		}
		else if (stricmp(subject, "DoRebar") == 0) {
			DoRebar((HWND) data1);
		}
		else if (stricmp(subject, "GetList") == 0) {
			DoGetList((char **) data2);
		}
		else if (stricmp(subject, "SwitchTab") == 0 ||
				 stricmp(subject, "CloseTab") == 0 ||
				 stricmp(subject, "DestroyTab") == 0) {
			PostUpdateMessage();
		}
		else return 0;
		return 1;
	}
	return 0;
}


void Load()
{
	g_id = kPlugin.kFuncs->GetCommandIDs(1);
}


void Create(HWND hWndParent)
{
	/* Should use SetWindowLongPtr */
	g_KMeleonWndProc = (WNDPROC) SetWindowLong(hWndParent, GWL_WNDPROC,
	                                           (LONG) WndProc);
}


void Close(HWND hWnd)
{
	HWND hwndRebar = FindWindowEx(hWnd, NULL, REBARCLASSNAME, NULL);
	if (! hwndRebar) return;
	HWND hwndIPBar = (HWND) RemoveProp(hwndRebar, BARPROPNAME);
}


void DoRebar(HWND rebarWnd)
{
	HWND hwndChild = CreateWindowEx(
		0,
		WC_COMBOBOXEX, TEXT(""),
		WS_CHILD | CBS_DROPDOWNLIST,
		0, 0, 0, 128, rebarWnd, NULL,
		kPlugin.hDllInstance, NULL);
	if (hwndChild == NULL) return;
	SendMessage(hwndChild, CBEM_SETEXSTYLE, 0, 
		CBES_EX_NOEDITIMAGE | CBES_EX_NOEDITIMAGEINDENT);
	
	HFONT hFont;
	hFont = (HFONT) SendMessage(rebarWnd, WM_GETFONT, 0, 0);
	if (hFont != NULL)
		SendMessage(hwndChild, WM_SETFONT, (WPARAM) hFont, TRUE);
	
	HDC dc = GetDC(hwndChild);
	static const LPCTSTR szIP = TEXT("000.000.000.000");
	static const int min = 3 * 4 + 3;
	SIZE size = {16 * min, 16};
	if (dc) {
		GetTextExtentPoint32(dc, szIP, min, &size);
		ReleaseDC(hwndChild, dc);
	}
	
	kPlugin.kFuncs->RegisterBand(hwndChild, TOOLBAND_NAME, TRUE);
	
	REBARBANDINFO rbi = {0};
	rbi.cbSize = sizeof(REBARBANDINFO);
	rbi.fMask  = RBBIM_TEXT | RBBIM_STYLE | RBBIM_CHILD |
		RBBIM_CHILDSIZE | RBBIM_SIZE | RBBIM_IDEALSIZE;
	rbi.lpText = TEXT("IP");
	rbi.fStyle = RBBS_CHILDEDGE | RBBS_FIXEDBMP | RBBS_VARIABLEHEIGHT;
	rbi.hwndChild  = hwndChild;
	rbi.cxMinChild = size.cx;
	rbi.cyMinChild = size.cy + GetSystemMetrics(SM_CXEDGE);
	rbi.cyChild = rbi.cyMinChild;
	rbi.cyMaxChild = rbi.cyMinChild;
	rbi.cyIntegral = 0;
	rbi.cx = rbi.cxMinChild;
	SendMessage(rebarWnd, RB_INSERTBAND, (WPARAM) -1, (LPARAM) &rbi);
	SetProp(rebarWnd, BARPROPNAME, hwndChild);
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message,
	WPARAM wParam, LPARAM lParam)
{
	switch (message) {
	case UWM_UPDATEBUSYSTATE:
		switch (wParam) {
		case 0:
			DoUpdate(hWnd);
			break;
		}
		break;
	case WM_COMMAND:
		if (LOWORD(wParam) == g_id) {
			DoUpdate(hWnd);
		}
		break;
	case WM_CLOSETAB:
	case WM_SWITCHTAB:
		DoUpdate(hWnd);
		break;
	}
	return CallWindowProc(g_KMeleonWndProc, hWnd, message, wParam, lParam);
}


extern "C" {
	BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
	{
		switch (dwReason) {
		case DLL_PROCESS_DETACH:
			break;
		}
		return TRUE;
	}
	
	KMELEON_PLUGIN kmeleonPlugin *GetKmeleonPlugin() {
		return &kPlugin;
	}
	
	KMELEON_PLUGIN int DrawBitmap(DRAWITEMSTRUCT *dis) {
		return 14; // 14 = icon width
	}
}


int message_box(HWND hwnd, LPCSTR lpText, LPCSTR lpTitle, UINT style)
{
	if (IsWindowUnicode(hwnd)) {
		return MessageBoxW(hwnd, miniWString(kPlugin.kFuncs->Translate(lpText)), miniWString(kPlugin.kFuncs->Translate(lpTitle)), style);
	}
	return MessageBoxA(hwnd, kPlugin.kFuncs->Translate(lpText), kPlugin.kFuncs->Translate(lpTitle), style);
}


HWND get_frame_hwnd()
{
	nsresult rv;
	
	miniCOMPtr<nsIWebBrowser> browser;
	if (! kPlugin.kFuncs->GetMozillaWebBrowser(NULL, &browser)) {
		return NULL;
	}
	
	miniCOMPtr<nsIWebBrowserChrome> chrome;
	rv = browser->GetContainerWindow(&chrome);
	NS_ENSURE_SUCCESS(rv, NULL);
	NS_ENSURE_TRUE(chrome, NULL);
	
	miniCOMPtr<nsIEmbeddingSiteWindow> esw;
	rv = chrome.QueryInterface(esw);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(esw, NULL);
	
	HWND hwnd = NULL;
	esw->GetSiteWindow((void **) &hwnd);
	
	return hwnd;
}


void PostUpdateMessage()
{
	HWND hwnd = get_frame_hwnd();
	if (hwnd) {
		PostMessage(hwnd, WM_COMMAND, g_id, 0);
	}
}


BOOL get_hostname(HWND hwnd, nsEmbedCString& hostname)
{
	nsresult rv;
	
	miniCOMPtr<nsIWebBrowser> browser;
	if (! kPlugin.kFuncs->GetMozillaWebBrowser(hwnd, &browser)) {
		return FALSE;
	}
	
	miniCOMPtr<nsIWebNavigation> nav;
	rv = browser.QueryInterface(nav);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(nav, FALSE);
	
	miniCOMPtr<nsIURI> uri;
	rv = nav->GetCurrentURI(&uri);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(uri, FALSE);
	
	rv = uri->GetHost(hostname);
	NS_ENSURE_SUCCESS(rv, FALSE);
	
	return TRUE;
}


HWND get_ipbar_hwnd(HWND hwnd)
{
	HWND hwndRebar, hwndChild;
	hwndRebar = hwndChild = NULL;
	hwndRebar = FindWindowEx(hwnd, NULL, REBARCLASSNAME, NULL);
	if (hwndRebar) {
		hwndChild = (HWND) GetProp(hwndRebar, BARPROPNAME);
	}
	return hwndChild;
}


class CDNSListener : public nsIDNSListener
{
public:
	NS_DECL_ISUPPORTS;
	NS_DECL_NSIDNSLISTENER;
	
	CDNSListener(HWND hwnd) : m_hwnd(hwnd) {}
	virtual ~CDNSListener() { }
private:
	CDNSListener() {}
protected:
	HWND m_hwnd;
};

NS_IMPL_ISUPPORTS1(CDNSListener, nsIDNSListener)

NS_IMETHODIMP CDNSListener::OnLookupComplete(nsICancelable *aRequest,
                                             nsIDNSRecord *aRecord,
                                             nsresult aStatus)
{
	NS_ENSURE_SUCCESS(aStatus, NS_ERROR_FAILURE);
	nsEmbedCString ip;
	nsresult rv;
	PRBool has;
	SendMessage(m_hwnd, CB_RESETCONTENT, 0, 0);
	for (;;) {
		rv = aRecord->HasMore(&has);
		NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
		if (! has) break;
		rv = aRecord->GetNextAddrAsString(ip);
		NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
		COMBOBOXEXITEM cbei = {0};
		cbei.mask = CBEIF_TEXT;
		cbei.iItem = -1;
		cbei.pszText = (char *) ip.get();
		SendMessage(m_hwnd, CBEM_INSERTITEM, 0, (LPARAM) &cbei);
	}
	SendMessage(m_hwnd, CB_SETCURSEL, 0, 0);
	return NS_OK;
}


BOOL DoUpdate(HWND hwnd)
{
	if (! hwnd) return FALSE;
	HWND hwndIPBar = get_ipbar_hwnd(hwnd);
	if (! hwndIPBar) return FALSE;
	
	SendMessage(hwndIPBar, CB_RESETCONTENT, 0, 0);
	
	int n = 0;
	kPlugin.kFuncs->GetPreference(PREF_INT, "network.proxy.type", &n, &n);
	if (n != 0) return FALSE;
	kPlugin.kFuncs->GetPreference(PREF_INT, "kmeleon.general.offline", &n, &n);
	if (n != 0) return FALSE;
	
	nsEmbedCString hostname;
	if (! get_hostname(hwnd, hostname) || hostname.Length() == 0)
		return FALSE;
	
	nsresult rv;
	miniCOMPtr<nsIDNSService> dnsservice;
	rv = dnsservice.doGetService(NS_DNSSERVICE_CONTRACTID);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(dnsservice, FALSE);
	
	miniCOMPtr<CDNSListener> dnslistener(new CDNSListener(hwndIPBar));
	miniCOMPtr<nsICancelable> cancel;
	rv = dnsservice->AsyncResolve(hostname, 0, dnslistener, nsnull, &cancel);
	NS_ENSURE_SUCCESS(rv, FALSE);
	
	return TRUE;
}


void DoGetList(char **result)
{
	HWND hwnd, hwndIPBar;
	nsEmbedCString list;
	if ((hwnd = get_frame_hwnd()) && (hwndIPBar = get_ipbar_hwnd(hwnd))) {
		static const int BUFFER_LENGTH = 256; /* IP is shorter than 256. */
		CHAR buffer[BUFFER_LENGTH];
		for (int i = 0; ; i ++) {
			/* Accoring to the MSDN, after sending CBEM_GETITEM,
			  pszText member of COMBOBOXEXITEM may be changed to
			  point another. Then this initialize on every loop. */
			COMBOBOXEXITEM cbei = {0};
			cbei.mask = CBEIF_TEXT;
			cbei.iItem = i;
			cbei.pszText = buffer;
			cbei.cchTextMax = BUFFER_LENGTH;
			if (! SendMessage(hwndIPBar, CBEM_GETITEM, 0, (LPARAM) &cbei))
				break;
			/* Append (+=) function doesn't return error state,
			   then this assumes success to all. */
			if (i > 0) list += "\r\n";
			list += cbei.pszText; /* Do not use "buffer". */
		}
	}
	*result = _strdup(list.get());
}
